//
// Created by daniel on 21.5.20.
//

#ifndef PROGRAM_COMMAND_H
#define PROGRAM_COMMAND_H


/*
 * --------------------------------------------
 * Blif file - Command
 * --------------------------------------------
 *
 * command is one of: <model-reference>
 *                    <clock-constraint>
 *                    <generic-latch>
 *                    <subfile-reference>
 *                    <delay-constraint>
 *                    <logic-gate>
 *                    <library-gate>
 *                    <fsm-description>
 *
 *
 */

#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <list>
#include <algorithm>
#include <memory>
#include <functional>

#include "commandtype.h"
#include "../format.h"
#include "utils/simulationtree.h"

class Command : public std::enable_shared_from_this<Command>, public Format {
public:
    Command(const CommandType &type)
    : Format(""),
      m_Type(type)
    {
    }

    Command(const std::string & name, const CommandType &type)
    : Format(name),
      m_Type(type)
    {
    }

    virtual void AddSimunation(const std::string &input, const std::string &output) {
        m_SimulationTree.Insert(input);
    }

    char GetValue() {
        if ( !m_Counted ) {
            this->m_Counted = true;
            std::string sequence;
            for(size_t i=0; i<m_Inputs.size(); ++i)
                sequence+=m_Inputs[i]->GetValue();
            SetValue( m_SimulationTree.Find(sequence) );
        }
        return m_Val;
    }

    void SetValue(const char val = '0')
    {
        if ( val != '0' && val!= '1' && val!='-' )
            throw "Invalid input: SetValue.";
        m_Val = val;
    }

    virtual std::shared_ptr<Command> Clone () const = 0;

    virtual void Print (std::ostream & os, const std::string & newLinePrefix, bool lastLine) {
        std::string name = this->m_Name;


        if (this->GetType() == LATCH)
            name = "{" + name + "}";

        if (this->isOutput())
            name = "(" + name + ")";

        if( this->isInput() )
        {
            name = "[" + name + "]";
            os << "-" <<  name << std::endl;
            if (!lastLine)
                os << newLinePrefix;
        }
        else if ( this->m_Visited )
        {
            os << "-" <<  name <<  std::endl;
            if (!lastLine)
                os << newLinePrefix;
        }
        else
        {
            this->m_Visited = true;
            size_t size = m_Inputs.size();
            std::string prefix = " ";
            for(size_t i=0; i<name.size(); ++i)
                prefix += " ";

            if ( size == 1 ) {
                os << "-" << name;
                m_Inputs[0]->Print(os, newLinePrefix + prefix, true);
            } else {
                os << "-" << name << "+";
                for(size_t i=0; i<size-2; ++i)
                    m_Inputs[i]->Print(os, newLinePrefix + prefix + "|", false);
                m_Inputs[size-2]->Print(os, newLinePrefix + prefix + "|", true);
                os << newLinePrefix << prefix << "\\";
                m_Inputs[size-1]->Print(os, newLinePrefix + prefix + " ", true);
            }
            if (!lastLine)
                os << newLinePrefix;
        }
    }

    void addInput(Command * input) {
        m_Inputs.push_back(input);
    }

    void addOutput(Command * output) {
        m_Outputs.push_back(output);
    }

    bool isInput() const {
        return m_Inputs.empty();
    }

    bool isOutput() const {
        return m_Output;
    }

    CommandType GetType() const {
        return m_Type;
    }


    std::vector<Command*> GetInputs() const {
        return m_Inputs;
    };

    std::string GetDataSimulation() const {
        return m_SimulationTree.GetData();
    }

    void PrintSimulation() const {
        m_SimulationTree.Print();
    }


public:
    bool m_Visited = false;
    bool m_Counted = false;
    bool m_Output = false;
protected:
    CommandType m_Type;
    std::vector<Command*> m_Inputs;
    std::vector<Command*> m_Outputs;
    SimulationTree m_SimulationTree;
    char m_Val;
};




#endif //PROGRAM_COMMAND_H
