//
// Created by daniel on 21.5.20.
//

#ifndef PROGRAM_MODEL_H
#define PROGRAM_MODEL_H


/*
 * --------------------------------------------
 * Blif file - Model
 * --------------------------------------------
 *
 * .model <decl-model-name>
 * .inputs <decl-input-list>
 * .outputs <decl-output-list>
 * .clock <decl-clock-list>
 * <command>
 * .
 * .
 * .
 * <command>
 * .end
 *
 * decl-model-name is a string giving the name of the model.
 * decl-input-list is a white-space-separated list of strings (terminated by the end of the line) giving the formal
 * input terminals for the model being declared. If this is the first or only model, then these signals can be
 * identified as the primary inputs of the circuit. Multiple .inputs lines are allowed, and the lists of inputs are
 * concatenated.
 *
 * decl-output-list is a white-space-separated list of strings (terminated by the end of the line) giving the formal
 * output terminals for the model being declared. If this is the first or only model, then these signals can be
 * identified as the primary outputs of the circuit. Multiple .outputs lines are allowed, and the lists of outputs
 * are concatenated.
 *
 * decl-clock-list is a white-space-separated list of strings (terminated by the end of the line) giving the clocks for
 * the model being declared. Multiple .clock lines are allowed, and the lists of clocks are concatenated.
 *
 * command is one of: <model-reference>
 *                    <clock-constraint>
 *                    <generic-latch>
 *                    <subfile-reference>
 *                    <delay-constraint>
 *                    <logic-gate>
 *                    <library-gate>
 *                    <fsm-description>
 *
 *
 */

#include <fstream>

#include "format.h"
#include "clock.h"
#include "command/command.h"
#include "command/used/latch.h"
#include "command/used/logicgate.h"


class Model : public Format, public std::enable_shared_from_this<Model> {
public:
    Model(const std::string &name = "")
            : Format(name) {}

    ~Model() {
        for (auto &command : m_Commands)
            delete (command.second);
    }

    Command *GetCommand(const std::string &name) {
        return (m_Commands.find(name) == m_Commands.end()) ? nullptr : m_Commands[name];
    }

    void Start() {
        this->Clear();
    }

    void SetOutput(const std::string & name) {
        if ( m_Commands.find(name)!=m_Commands.end() && !m_Commands[name]->isOutput() )
        {
            m_Commands[name]->m_Output = true;
            m_Outputs.push_back(m_Commands[name]);
        }
    }

    void Stop() {
        FindComponents();
        for(auto const & it: m_Inputs)
            m_AllLatches.push_back( it );
    }

    void AddCommand(Command  * command) {
        if ( m_Commands.find(command->GetName())== m_Commands.end() ) {
            m_Commands.insert( {command->GetName(), command } );
        } else {
            delete(command);
        }
    }

    void AddConnection(const std::string &input, const std::string &output) {
        AddCommand(new LogicGate(input));
        AddCommand(new LogicGate(output));
        m_Commands[output]->addInput( m_Commands[input] );
        m_Commands[input]->addOutput( m_Commands[output] );
    }



    bool Simulation(const std::vector<std::string> & inputs, std::vector<std::string> & outputs)
    {
        for(auto const & input : inputs)
        {
        	if ( input.size() != m_Inputs.size() )
        	{
        		outputs.push_back("error");
        		continue;
        	}
            
            std::string output = SimulationOne(input);
            if (!output.empty())
            	outputs.push_back(output);
        }
    	return true;
    }

    std::map<std::string, Command*> GetCommands() const {
        return m_Commands;
    }

    std::vector<Command*> GetLatches() const {
        return m_AllLatches;
    }
    std::vector<Command*> GetOutputs() const  {
        return m_Outputs;
    }

    std::vector<Command*> GetInputs() const {
        return m_Inputs;
    }


    bool IsEmpty() const
    {
        return m_Commands.empty();
    }

private:
    std::string SimulationOne(const std::string &input)
    {
        std::string output;
        if ( m_Inputs.size() != input.size() )
            return "";

        for(auto & command : m_Commands)
            command.second->m_Counted = false;

        for(size_t i=0; i<m_Inputs.size(); ++i)
        {
            m_Inputs[i]->m_Counted = true;
            m_Inputs[i]->SetValue(input[i]);
        }

        for(size_t i=0; i<m_Outputs.size(); ++i)
            output += m_Outputs[i]->GetValue();
        return output;
    }

    void Clear()
    {
        m_Outputs.clear();
        m_Inputs.clear();
        m_Commands.clear();
        m_AllLatches.clear();
    }

    void FindComponents()
    {
        m_Inputs.clear();
        for(auto const &  command : m_Commands) {
            if ( command.second->isInput() )
                m_Inputs.push_back( command.second );
        }
    }


    void Print(std::ostream &os) const
    {
        os.clear();
        if (this->IsEmpty())
        {
            os <<  "Empty model: [" << this->GetName() << "]";
            return;
        }

        os <<  "Model: [" << this->GetName() << "]" << std::endl;

        for(auto const & it : this->m_Commands)
            it.second->m_Visited = false;


        std::string prefix = "";
        os <<this->m_Name;
        for(size_t i=0; i<this->m_Name.size(); ++i)
            prefix += " ";

        size_t size = this->m_Outputs.size();


        if ( size == 1 ) {
            this->m_Outputs[0]->Print(os, prefix, true);
        } else if ( size > 1 ) {
            os << "+";
            for(size_t i=0; i<size-2; ++i)
                this->m_Outputs[i]->Print(os, prefix +  "|", false);
            this->m_Outputs[size-2]->Print(os, prefix + "|", true);
            os << prefix << "\\";
            this->m_Outputs[size-1]->Print(os, prefix + " ", true);
        } else {
            os << ": Empty model!!!" << std::endl ;
        }

    }



    void PrintAll() const {
        std::cout << "--------" << std::endl;
        for(auto const & command : m_Commands) {
            std::cout << command.first << std::endl;
            command.second->PrintSimulation();
            std::cout << "-" << std::endl;
        }
    }

    void PrintLatches(std::ostream & os) const
    {
        os << "-------- Latches --------" << std::endl;
        for(auto const & it : this->m_AllLatches)
            os << "-" << "{" << it->GetName() << "}";

        os << "-" << std::endl;
    }


    friend std::ostream & operator << (std::ostream &os, const Model &model) {
        model.Print(os);
        return os;
    }

private:


    std::map<std::string, Command*> m_Commands;
    std::vector<Command*> m_Outputs;
    std::vector<Command*> m_Inputs;
    std::vector<Command*> m_AllLatches;
};
typedef std::unique_ptr<Model> UModel;

#endif //PROGRAM_MODEL_H
